/*****************************


******************************/
#include "zoomcam.h"
#include "camera.h"

#include "target.h"
#include "macrorec.h"
#include "modstack.h"
#include "iFnPub.h"
#include "bmmlib.h"

#include "stdmat.h"

#define ZOOMCAMERA_FP_INTERFACE_ID Interface_ID(0x438a1133, 0xef977711)

class ZoomCameraFunctionPublish : public FPStaticInterface {
	public:
		DECLARE_DESCRIPTOR(ZoomCameraFunctionPublish);

		enum OpID {
			kCreate
			
			};
		
		BEGIN_FUNCTION_MAP			
			VFN_0(kCreate,  Create)

		END_FUNCTION_MAP

		
		void Create();
	};

static ZoomCameraFunctionPublish theZoomCameraFunctionPublish(
	ZOOMCAMERA_FP_INTERFACE_ID, _T("theZoomCamera"), -1, 0, FP_CORE,
	// The first operation, boneCreate:
	ZoomCameraFunctionPublish::kCreate, _T("Create"), -1, TYPE_VOID, 0, 0,

	end);



void ZoomCameraFunctionPublish::Create()
{
	BOOL cameraCreated = FALSE;

	Interface *ip = GetCOREInterface();

	if (ip)

	{

	//get the selected object
		INode *cameraNode = NULL;
	//if camera
		int ct = ip->GetSelNodeCount();

		for (int i = 0; i < ct; i++)
		{
			cameraNode = ip->GetSelNode(i);

			Object *cameraObj = cameraNode->GetObjectRef();



			if (cameraObj->SuperClassID() == CAMERA_CLASS_ID)
			{
				i = ct;
				Object *obj = (Object*)ip->CreateInstance(CAMERA_CLASS_ID,	Class_ID(ZOOMSIMPLE_CAM_CLASS_ID, 0));

				SimpleZoomCamera *zcamera = (SimpleZoomCamera *)obj;
				
				zcamera->pblock->SetValue(camera_param_basecam,0,cameraNode);
				INode *node = ip->CreateObjectNode(obj);
				node->SetName("ZoomCamera");
			// Create child node
				NullPosition *nullCtrl;
				nullCtrl = (NullPosition *)	CreateInstance(CTRL_MATRIX3_CLASS_ID,Class_ID(NULL_CONTROL_CLASS_ID,0)); 			
				
				cameraCreated = TRUE;

				float zoom = ((CameraObject*) cameraObj)->GetFOV(0,FOREVER);

				if (nullCtrl)
					{
		
					nullCtrl->pblock->SetValue(null_param_zoom,0,zoom);
					node->SetTMController(nullCtrl);
					nullCtrl->pblock->SetValue(null_params_backpointer,0,(ReferenceTarget*)obj);					
					}

				//create our image plnae node
  				Object *gridObj = (Object*)ip->CreateInstance(GEOMOBJECT_CLASS_ID,	Class_ID(GRIDID));
				
				INode *gridNode = ip->CreateObjectNode(gridObj);
				gridNode->SetName("ZoomCamera_ImagePlane");

				//create our image plane
				//add the null ctrl
				NullPosition *gnullCtrl;
				gnullCtrl = (NullPosition *)	CreateInstance(CTRL_MATRIX3_CLASS_ID,Class_ID(NULL_CONTROL_CLASS_ID,0)); 			
				
				GridObject *gobj = (GridObject *) gridObj;
				gobj->pblock2->SetValue(grid_camera,0,cameraNode);

				if (gnullCtrl)
					{
		

					Control *c = NewDefaultFloatController();
					ParamID pid;

					pid = null_param_zoffset;
					gnullCtrl->pblock->SetController(pid,0,c);
					pid = grid_zdepth;
					gobj->pblock2->SetController(pid,0,c);


 					c = NewDefaultFloatController();
					pid = null_param_panx;
					nullCtrl->pblock->SetController(pid,0,c);
					pid = grid_panx;
					gobj->pblock2->SetController(pid,0,c);

					c = NewDefaultFloatController();
					pid = null_param_pany;
					nullCtrl->pblock->SetController(pid,0,c);
					pid = grid_pany;
					gobj->pblock2->SetController(pid,0,c);


					c = NewDefaultFloatController();
					pid = null_param_zoom;
					nullCtrl->pblock->SetController(pid,0,c);
					pid = grid_fov;
					gobj->pblock2->SetController(pid,0,c);

					c = NewDefaultBoolController();
					pid = null_param_animatable;
					nullCtrl->pblock->SetController(pid,0,c);
					pid = grid_animatable;
					gobj->pblock2->SetController(pid,0,c);

					gnullCtrl->pblock->SetValue(null_param_zoffset,0,-500.0f);
					nullCtrl->pblock->SetValue(null_param_zoom,0,zoom);
					nullCtrl->pblock->SetValue(null_param_panx,0,0.0f);
					nullCtrl->pblock->SetValue(null_param_pany,0,0.0f);
//					ResumeAnimate();
//					ParamID pid;
//					pid = null_param_zoffset;
//					Control *c = gnullCtrl->pblock->GetController(null_param_zoffset);

					gridNode->SetTMController(gnullCtrl);
					gnullCtrl->pblock->SetValue(null_param_zoom,0,zoom);
					gnullCtrl->pblock->SetValue(null_params_backpointer,0,(ReferenceTarget*)obj);			


					StdMat *stdMat = NewDefaultStdMat();
					stdMat->SetName("UnwrapChecker");					
					BitmapTex *bitmapMap = (BitmapTex *) CreateInstance(TEXMAP_CLASS_ID, Class_ID(BMTEX_CLASS_ID,0));
					StdUVGen *uvGen = (StdUVGen*)bitmapMap->GetTheUVGen();
					uvGen->SetUScl(1.0f, 0);
					uvGen->SetVScl(1.0f, 0);
					bitmapMap->BitmapLoadDlg();
					bitmapMap->SetName("BitmapPattern");
					stdMat->SetSubTexmap(1, bitmapMap);
					
					gridNode->SetMtl(stdMat);

					ip->ActivateTexture(stdMat->GetSubTexmap(1), stdMat, 1);


					}


			}
		}
	}
	
 	if (!cameraCreated)
	{
		MessageBox(ip->GetMAXHWnd(),"Error no camera selected.\nMake sure to select a camera to apply the zoom camera to before running the macro","Error No Camera Selected",MB_OK);
	}

}

