/**********************************************************************
 *<
	FILE: camera.cpp

	DESCRIPTION:  A Simple Camera implementation

	CREATED BY: Dan Silva

	HISTORY: created 14 December 1994

	add the  z depth control to the null control
	add the dialog to the grid
	add the camera node
	add the null ctrl to the grid

	privatize the grid

	add the action item
		add bitmap controls
		hookup to fov


	add a control dialog



 *>	Copyright (c) 1994, All Rights Reserved.
 **********************************************************************/

#include "zoomcam.h"
#include "gencam.h"
#include "camera.h"
#include "target.h"
#include "macrorec.h"
#include "modstack.h"

extern void BuildCameraMesh(Mesh *mesh);

#define WM_SET_TYPE		WM_USER + 0x04002

#define NUM_CIRC_PTS	28
#define SEG_INDEX		7

static HIMAGELIST hCamImages = NULL;

//------------------------------------------------------
class SimpleCamClassDesc:public ClassDesc2 {
	public:
	int 			IsPublic() { return 0; }
	void *			Create(BOOL loading = FALSE) { return new SimpleZoomCamera; }
	const TCHAR *	ClassName() { return GetString(IDS_DB_FREE_CLASS); }
	SClass_ID		SuperClassID() { return CAMERA_CLASS_ID; }
	Class_ID 		ClassID() { return Class_ID(ZOOMSIMPLE_CAM_CLASS_ID,0); }
	const TCHAR* 	Category() { return _T("");  }
//	void			ResetClassParams(BOOL fileReset) { if(fileReset) resetCameraParams(); }

	const TCHAR*	InternalName() { return _T("ZoomCamera"); }	// returns fixed parsable name (scripter-visible name)
	HINSTANCE		HInstance() { return hInstance; }			// returns owning module handle

	};

static SimpleCamClassDesc simpleCamDesc;

ClassDesc* GetSimpleCamDesc() { return &simpleCamDesc; }





// class type-in block
static ParamBlockDesc2 camera_blk ( camera_param, _T("Parameters"),  0, &simpleCamDesc, P_AUTO_CONSTRUCT + P_AUTO_UI, 0, 
	//rollout
	IDD_SCAMERAPARAM, IDS_RB_PARAMETERS, 0, 0, NULL,


	camera_param_basecam,  _T("camera"), 			TYPE_INODE, P_RESET_DEFAULT, 	IDS_PW_CAMERA, 
		p_ui, 			TYPE_PICKNODEBUTTON, IDC_CAMERA, 
		end,

	camera_mtl,		 _T("map"),	TYPE_BITMAP, 0, IDS_PW_CAMERA,
		p_ui,			TYPE_BITMAPBUTTON, IDC_MATERIAL,
		end,
/*
	camera_param_enable,	_T("enablePlane"), TYPE_BOOL_TAB, 4, 0, IDS_PLANEENABLE,
		p_default,		FALSE,
		p_ui,			TYPE_SINGLECHEKBOX, IDC_CHECK1, IDC_CHECK2, IDC_CHECK3, IDC_CHECK4,
											
		end,

	camera_param_depth,  _T("zamount"),		TYPE_FLOAT_TAB, 	4,P_ANIMATABLE|P_TV_SHOW_ALL,	IDS_PW_AMOUNT, 
		p_default, 		100.0f,	
		p_range, 		0.00f, 2000000.0f, 
		p_ui, 			TYPE_SPINNER, EDITTYPE_FLOAT, 
						IDC_PLANEZ1,IDC_PLANEZSPINNER1,
						IDC_PLANEZ2,IDC_PLANEZSPINNER2,
						IDC_PLANEZ3,IDC_PLANEZSPINNER3,
						IDC_PLANEZ4,IDC_PLANEZSPINNER4,
//						IDC_COMPMAP_AMOUNT10,IDC_COMPMAP_AMOUNTSPIN10,
						1.0f, //SPIN_AUTOSCALE, 
		end, 

*/
/*
	camera_param_zoom,  _T("zoom"), 			TYPE_FLOAT, 	P_ANIMATABLE + P_RESET_DEFAULT, 	IDS_PW_ZOOM, 
		p_default, 		1.0f,	
		p_ms_default,	1.0f,
		p_range, 		0.0f, 10.f, 
		p_ui, 			TYPE_SPINNER, EDITTYPE_UNIVERSE, IDC_ZOOM, IDC_ZOOMSPINNER, 0.1f, 
		end, 

	camera_param_panx,  _T("pan_x"), 			TYPE_FLOAT, 	P_ANIMATABLE + P_RESET_DEFAULT, 	IDS_PW_PANX, 
		p_default, 		0.0f,	
		p_ms_default,	0.0f,
		p_range, 		-100.0f, 100.0f, 
		p_ui, 			TYPE_SPINNER, EDITTYPE_UNIVERSE, IDC_PANX, IDC_PANXSPINNER, 1.f, 
		end, 

	camera_param_pany,  _T("pan_y"), 			TYPE_FLOAT, 	P_ANIMATABLE + P_RESET_DEFAULT, 	IDS_PW_PANY, 
		p_default, 		0.0f,	
		p_ms_default,	0.0f,
		p_range, 		-100.0f, 100.0f, 
		p_ui, 			TYPE_SPINNER, EDITTYPE_UNIVERSE, IDC_PANY, IDC_PANYSPINNER, 1.f, 
		end, 
*/
	end
	);

// in prim.cpp  - The dll instance handle
extern HINSTANCE hInstance;

#define DEF_HITHER_CLIP		1.0f
#define DEF_YON_CLIP		1000.0f

// Class variables of SimpleZoomCamera
Mesh SimpleZoomCamera::mesh;
short SimpleZoomCamera::meshBuilt=0;


BOOL SimpleZoomCamera::inCreate=FALSE;


HWND SimpleZoomCamera::hSimpleCamParams = NULL;
IObjParam *SimpleZoomCamera::iObjParams;



static float GetAspect() {
	return GetCOREInterface()->GetRendImageAspect();
	};

static float GetApertureWidth() {
	return GetCOREInterface()->GetRendApertureWidth();
	}


static int GetTargetPoint(TimeValue t, INode *inode, Point3& p) 
	{
	return 0;
	}

static int is_between(float a, float b, float c) 
{
	float t;
	if (b>c) { t = b; b = c; c = t;}
	return((a>=b)&&(a<=c));
}

static float interp_vals(float m, float *mtab, float *ntab, int n) 
{
	float frac;
	for (int i=1; i<n; i++) {
		if (is_between(m,mtab[i-1],mtab[i])) {
			frac = (m - mtab[i-1])/(mtab[i]-mtab[i-1]);
			return((1.0f-frac)*ntab[i-1] + frac*ntab[i]);
		}
	}
	return 0.0f;
}



class CameraDlgProc : public ParamMap2UserDlgProc {
public:
	SimpleZoomCamera *mod;		
	CameraDlgProc(SimpleZoomCamera *m) {mod = m;}		
	BOOL DlgProc(TimeValue t,IParamMap2 *map,HWND hWnd,UINT msg,WPARAM wParam,LPARAM lParam);		
	void DeleteThis() {delete this;}		
};

BOOL CameraDlgProc::DlgProc (TimeValue t,IParamMap2 *map,HWND hWnd,UINT msg,WPARAM wParam,LPARAM lParam)
	{
	switch (msg) {		
		case WM_COMMAND:
			switch (LOWORD(wParam)) {

				case IDC_RESET:
					mod->ResetParams();
					break;

		
				}
			break;
		}
	return FALSE;
	}

void SimpleZoomCamera::ResetParams()
	{
	if (nullCtrl)
		{
		float zoom = 1.0f;
		if (baseCam) zoom = baseCam->GetFOV(0,FOREVER);
		nullCtrl->pblock->SetValue(null_param_zoom,0,zoom);
		nullCtrl->pblock->SetValue(null_param_panx,0,0.0f);
		nullCtrl->pblock->SetValue(null_param_pany,0,0.0f);
		}
	}


// This changes the relation between FOV and Focal length.
void SimpleZoomCamera::RenderApertureChanged(TimeValue t) 
	{
	if (baseCam) baseCam->RenderApertureChanged(t) ;
	}






static void LoadCamResources() {
	static BOOL loaded=FALSE;
	if (loaded) return;
	HBITMAP hBitmap;
	HBITMAP hMask;
	hCamImages = ImageList_Create(14,14, ILC_COLOR4|ILC_MASK, 3, 0);
	hBitmap = LoadBitmap(hInstance,MAKEINTRESOURCE(IDB_FOV));
	hMask   = LoadBitmap(hInstance,MAKEINTRESOURCE(IDB_FOVMASK));
	ImageList_Add(hCamImages,hBitmap,hMask);
	DeleteObject(hBitmap);	
	DeleteObject(hMask);	
	}

class DeleteCamResources {
	public:
		~DeleteCamResources() {
			ImageList_Destroy(hCamImages);
			}
	};

static DeleteCamResources theDelete;

static int typeName[NUM_CAM_TYPES] = {
	IDS_DB_FREE_CAM,
	IDS_DB_TARGET_CAM
	};



void SimpleZoomCamera::BeginEditParams( IObjParam *ip, ULONG flags,Animatable *prev )
{
	iObjParams = ip;	
	simpleCamDesc.BeginEditParams(ip, this, flags, prev);
	camera_blk.SetUserDlgProc(new CameraDlgProc(this));

}
		
void SimpleZoomCamera::EndEditParams( IObjParam *ip, ULONG flags,Animatable *prev)
{
	iObjParams = NULL;
	simpleCamDesc.EndEditParams(ip, this, flags, prev);
}


static void MakeQuad(Face *f, int a,  int b , int c , int d, int sg, int dv = 0) {
	f[0].setVerts( a+dv, b+dv, c+dv);
	f[0].setSmGroup(sg);
	f[0].setEdgeVisFlags(1,1,0);
	f[1].setVerts( c+dv, d+dv, a+dv);
	f[1].setSmGroup(sg);
	f[1].setEdgeVisFlags(1,1,0);
	}



void SimpleZoomCamera::BuildMesh()	{


	INode *node;
	pblock->GetValue(camera_param_basecam,0,node,FOREVER);

	if (node)
		{
		ObjectState os = node->EvalWorldState(0);
		if (os.obj->SuperClassID() == CAMERA_CLASS_ID)
			{
			baseCam = (CameraObject*) os.obj;

//			((GenCamera*)node->GetObjectRef())->Enable(TRUE);
			((GenCamera*)baseCam)->Enable(TRUE);

			}
		}

	BuildCameraMesh(&mesh);

	planeMesh.setNumVerts(4);
	planeMesh.setNumFaces(2);
	planeMesh.setMapSupport(1);
	planeMesh.setNumMapVerts(1, 4);
	planeMesh.setNumMapFaces(1, 2);

	UVVert *uvVerts = planeMesh.mapVerts(1);
	TVFace *uvFaces = planeMesh.mapFaces(1);
	
	uvVerts[0] = Point3(0.0f,0.0f,0.0f);
	uvVerts[1] = Point3(1.0f,0.0f,0.0f);
	uvVerts[2] = Point3(0.0f,1.0f,0.0f);
	uvVerts[3] = Point3(0.0f,1.0f,0.0f);

	uvFaces[0].t[0] = 0;
	uvFaces[0].t[1] = 1;
	uvFaces[0].t[2] = 2;

	uvFaces[1].t[0] = 2;
	uvFaces[1].t[1] = 1;
	uvFaces[1].t[2] = 3;



	planeMesh.setVert(0, Point3(-20.0f,-20.0f,0.0f));
	planeMesh.setVert(1, Point3(20.0f,-20.0f,0.0f));
	planeMesh.setVert(2, Point3(-20.0f,20.0f,0.0f));
	planeMesh.setVert(3, Point3(20.0f,20.0f,0.0f));

	Face f;
	f.v[0] = 0;
	f.v[1] = 1;
	f.v[2] = 2;
	f.smGroup = 0;
	f.flags = 0;
	f.setEdgeVisFlags(EDGE_VIS ,EDGE_VIS , EDGE_VIS );
	planeMesh.faces[0] = f;

	f.v[0] = 2;
	f.v[1] = 1;
	f.v[2] = 3;
	f.smGroup = 0;
	f.flags = 0;
	f.setEdgeVisFlags(EDGE_VIS ,EDGE_VIS , EDGE_VIS );
	planeMesh.faces[1] = f;


	


/*	int nverts = 16;
	int nfaces = 24;
	mesh.setNumVerts(nverts);
	mesh.setNumFaces(nfaces);
	float len = (float)5.0;
	float w = (float)8.0;
	float d = w*(float).8;
	float e = d*(float).5;
	float f = d*(float).8;
	float l = w*(float).8;

	mesh.setVert(0, Point3( -d, -d, -len));
	mesh.setVert(1, Point3(  d, -d, -len));
	mesh.setVert(2, Point3( -d,  d, -len));
	mesh.setVert(3, Point3(  d,  d, -len));
	mesh.setVert(4, Point3( -d, -d,  len));
	mesh.setVert(5, Point3(  d, -d,  len));
	mesh.setVert(6, Point3( -d,  d,  len));
	mesh.setVert(7, Point3(  d,  d,  len));
	MakeQuad(&(mesh.faces[ 0]), 0,2,3,1,  1);
	MakeQuad(&(mesh.faces[ 2]), 2,0,4,6,  2);
	MakeQuad(&(mesh.faces[ 4]), 3,2,6,7,  4);
	MakeQuad(&(mesh.faces[ 6]), 1,3,7,5,  8);
	MakeQuad(&(mesh.faces[ 8]), 0,1,5,4, 16);
	MakeQuad(&(mesh.faces[10]), 4,5,7,6, 32);
	
	mesh.setVert(8+0, Point3( -e, -e, len));
	mesh.setVert(8+1, Point3(  e, -e, len));
	mesh.setVert(8+2, Point3( -e,  e, len));
	mesh.setVert(8+3, Point3(  e,  e, len));
	mesh.setVert(8+4, Point3( -f, -f, len+l));
	mesh.setVert(8+5, Point3(  f, -f, len+l));
	mesh.setVert(8+6, Point3( -f,  f, len+l));
	mesh.setVert(8+7, Point3(  f,  f, len+l));

	Face* fbase = &mesh.faces[12];
	MakeQuad(&fbase[0],0,2,3,1,   1, 8);
	MakeQuad(&fbase[2], 2,0,4,6,  2, 8);
	MakeQuad(&fbase[4], 3,2,6,7,  4, 8);
	MakeQuad(&fbase[6], 1,3,7,5,  8, 8);
	MakeQuad(&fbase[8], 0,1,5,4, 16, 8);
	MakeQuad(&fbase[10],4,5,7,6, 32, 8);

	// whoops- rotate 180 about x to get it facing the right way
	Matrix3 mat;
	mat.IdentityMatrix();
	mat.RotateX(DegToRad(180.0));
	for (int i=0; i<nverts; i++)
		mesh.getVert(i) = mat*mesh.getVert(i);
*/
	mesh.buildNormals();
	mesh.EnableEdgeList(1);
	meshBuilt = 1;
	}




// mjm - begin - 07.17.00
SimpleZoomCamera::SimpleZoomCamera(int look)
{
initial = FALSE;
GetSimpleCamDesc()->MakeAutoParamBlocks(this); 
baseCam = NULL;
BuildMesh();
}






//--------------------------------------------

static INode* FindNodeRef(ReferenceTarget *rt);

static INode* GetNodeRef(ReferenceMaker *rm) {
	if (rm->SuperClassID()==BASENODE_CLASS_ID) return (INode *)rm;
	else return rm->IsRefTarget()?FindNodeRef((ReferenceTarget *)rm):NULL;
	}

static INode* FindNodeRef(ReferenceTarget *rt) {
	DependentIterator di(rt);
	ReferenceMaker *rm;
	INode *nd = NULL;
	while (rm=di.Next()) {	
		nd = GetNodeRef(rm);
		if (nd) return nd;
		}
	return NULL;
	}

void SimpleZoomCamera::SetType(int tp) {     
	}


//---------------------------------------
void SimpleZoomCamera::SetFOV(TimeValue t, float f) 
	{
	if (baseCam) 
		{
		float fov =  baseCam->GetFOV(t,FOREVER);
		float outfov = f/fov;
//		nullCtrl->pblock->SetValue(null_param_zoom,0,outfov);
		BOOL animatable;
		nullCtrl->pblock->GetValue(null_param_animatable,t,animatable,FOREVER);

		if (!animatable)
		{
			SuspendAnimate();
			AnimateOff();
		}

		nullCtrl->pblock->SetValue(null_param_zoom,t,f);
		if (!animatable)
		{
			ResumeAnimate();
		}

		}
	}

float SimpleZoomCamera::GetFOV(TimeValue t,Interval& valid) {	
	if (baseCam) 
		{
		float zoom = 1.0f;
		if (nullCtrl)
			{
			nullCtrl->pblock->GetValue(null_param_zoom,t,zoom,valid);
			}
//		return baseCam->GetFOV(t,valid)/zoom;
		return zoom;
		}
	return 0;
	}

void SimpleZoomCamera::SetTDist(TimeValue t, float f) 
	{
	if (baseCam) baseCam->SetTDist(t,f);

	}

float SimpleZoomCamera::GetTDist(TimeValue t,Interval& valid) {	
	if (baseCam) return baseCam->GetTDist(t,valid);
	return 0.0f;
}

void SimpleZoomCamera::SetManualClip(int onOff)
{
	if (baseCam) baseCam->SetManualClip(onOff);

}

float SimpleZoomCamera::GetClipDist(TimeValue t, int which, Interval &valid)
{
	if (baseCam) return baseCam->GetClipDist(t,which,valid);
	return 0.0f;
}

void SimpleZoomCamera::SetClipDist(TimeValue t, int which, float f)
{
	if (baseCam) baseCam->SetClipDist(t,which,f);

}


void SimpleZoomCamera::SetEnvRange(TimeValue t, int which, float f)
{
	if (baseCam) baseCam->SetEnvRange(t,which,f);

}

float SimpleZoomCamera::GetEnvRange(TimeValue t, int which, Interval &valid)
{
	float f=0.0f;
	if (baseCam) return baseCam->GetEnvRange(t,which,valid);

	return f;
}

void SimpleZoomCamera::SetEnvDisplay(BOOL b, int notify) {
	if (baseCam) baseCam->SetEnvDisplay(b, notify);
	}



// mjm - end


SimpleZoomCamera::~SimpleZoomCamera() {
	DeleteAllRefsFromMe();
	pblock = NULL;
	}

class SCamCreateCallBack: public CreateMouseCallBack {
	SimpleZoomCamera *ob;
	public:
		int proc( ViewExp *vpt,int msg, int point, int flags, IPoint2 m, Matrix3& mat );
		void SetObj(SimpleZoomCamera *obj) { ob = obj; }
	};

int SCamCreateCallBack::proc(ViewExp *vpt,int msg, int point, int flags, IPoint2 m, Matrix3& mat ) {
#ifdef _OSNAP
	if (msg == MOUSE_FREEMOVE)
	{
		#ifdef _3D_CREATE
			vpt->SnapPreview(m,m,NULL, SNAP_IN_3D);
		#else
			vpt->SnapPreview(m,m,NULL, SNAP_IN_PLANE);
		#endif
	}
#endif
	if (msg==MOUSE_POINT||msg==MOUSE_MOVE) {
		//mat[3] = vpt->GetPointOnCP(m);
#ifdef _3D_CREATE
		mat.SetTrans( vpt->SnapPoint(m,m,NULL,SNAP_IN_3D) );
#else
		mat.SetTrans( vpt->SnapPoint(m,m,NULL,SNAP_IN_PLANE) );
#endif
		if (point==1 && msg==MOUSE_POINT) 
			{
/*
			Object *obj = (Object*)GetCOREInterface()->CreateInstance(
				CAMERA_CLASS_ID,
				Class_ID(0x1001,0));
			assert(obj);
	// Name the node and make the name unique.
			IDerivedObject *dobj = CreateDerivedObject(obj);
			INode *node = GetCOREInterface()->CreateObjectNode(dobj);
			TSTR name(_T("Base Camera"));
			GetCOREInterface()->MakeNameUnique(name);
			node->SetName(name);
			node->SetNodeTM(0,mat);
			ob->pblock->SetValue(camera_param_basecam,0,node);
//			((GenCamera*)dobj)->Enable(TRUE);


			if (node)
				{
				ObjectState os = node->EvalWorldState(0);
				if (os.obj->SuperClassID() == CAMERA_CLASS_ID)
					{
					CameraObject *baseCam;
					baseCam = (CameraObject*) os.obj;

					((GenCamera*)baseCam)->Enable(TRUE);
	
					}
				}
*/
			return 0;
			}
		}
	else
	if (msg == MOUSE_ABORT)
		return CREATE_ABORT;

	return TRUE;
	}

static SCamCreateCallBack sCamCreateCB;

CreateMouseCallBack* SimpleZoomCamera::GetCreateMouseCallBack() {
	sCamCreateCB.SetObj(this);
	return(&sCamCreateCB);
	}


void SimpleZoomCamera::GetDeformBBox(TimeValue t, Box3& box, Matrix3 *tm, BOOL useSel )
	{
	box = mesh.getBoundingBox(tm);
	}

void SimpleZoomCamera::GetLocalBoundBox(TimeValue t, INode* inode, ViewExp* vpt, Box3& box )
	{
	Matrix3 m = inode->GetObjectTM(t);
	Point3 pt;
//	float scaleFactor = vpt->NonScalingObjectSize()*vpt->GetVPWorldWidth(m.GetTrans())/(float)360.0;
	box = mesh.getBoundingBox();

	}

void SimpleZoomCamera::GetWorldBoundBox(TimeValue t, INode* inode, ViewExp* vpt, Box3& box )
	{
	int i,nv;
	Matrix3 tm;
	Point3 pt;
	tm = inode->GetObjectTM(t);
	nv = mesh.getNumVerts();
	box.Init();
	for (i=0; i<nv; i++) 
		box += tm*mesh.getVert(i);

	tm = inode->GetObjectTM(t);

	}

void SimpleZoomCamera::UpdateTargDistance(TimeValue t, INode* inode) 
	{
	}

// From BaseObject
int SimpleZoomCamera::HitTest(TimeValue t, INode* inode, int type, int crossing, int flags, IPoint2 *p, ViewExp *vpt) {
	Point2 pt( (float)p[0].x, (float)p[0].y );
	HitRegion hitRegion;
	GraphicsWindow *gw = vpt->getGW();	
	
	gw->setTransform(inode->GetObjectTM(t));
	MakeHitRegion(hitRegion, type, crossing, 4, p);
	return mesh.select(gw, inode->Mtls(), &hitRegion, flags & HIT_ABORTONHIT, inode->NumMtls());
/*
	HitRegion hitRegion;
	DWORD	savedLimits;
	int res=0;
	Matrix3 m;
	GraphicsWindow *gw = vpt->getGW();	
	MakeHitRegion(hitRegion,type,crossing,4,p);	
	gw->setRndLimits(((savedLimits = gw->getRndLimits()) | GW_PICK) & ~GW_ILLUM);
	return res;
	*/
}

void SimpleZoomCamera::Snap(TimeValue t, INode* inode, SnapInfo *snap, IPoint2 *p, ViewExp *vpt) {
	// Make sure the vertex priority is active and at least as important as the best snap so far
	if(snap->vertPriority > 0 && snap->vertPriority <= snap->priority) {
		Matrix3 tm = inode->GetObjectTM(t);	
		GraphicsWindow *gw = vpt->getGW();	
   	
		gw->setTransform(tm);

		Point2 fp = Point2((float)p->x, (float)p->y);
		IPoint3 screen3;
		Point2 screen2;
		Point3 vert(0.0f,0.0f,0.0f);

		gw->wTransPoint(&vert,&screen3);

		screen2.x = (float)screen3.x;
		screen2.y = (float)screen3.y;

		// Are we within the snap radius?
		int len = (int)Length(screen2 - fp);
		if(len <= snap->strength) {
			// Is this priority better than the best so far?
			if(snap->vertPriority < snap->priority) {
				snap->priority = snap->vertPriority;
				snap->bestWorld = vert * tm;
				snap->bestScreen = screen2;
				snap->bestDist = len;
				}
			else // Closer than the best of this priority?
			if(len < snap->bestDist) {
				snap->priority = snap->vertPriority;
				snap->bestWorld = vert * tm;
				snap->bestScreen = screen2;
				snap->bestDist = len;
				}
			}
		}
	}

BITMAPINFO *BMToDIB(Bitmap* newBM, int extraFlags)	{
	int bmw = newBM->Width();
	int bmh = newBM->Height();
	BITMAPINFO* texBMI = (BITMAPINFO *)malloc(sizeof(BITMAPINFOHEADER) + bmw * bmh * sizeof(RGBQUAD));
	BITMAPINFOHEADER *bmih = &texBMI->bmiHeader;
	bmih->biSize = sizeof (BITMAPINFOHEADER);
	bmih->biWidth = bmw;
   	bmih->biHeight = bmh;
	bmih->biPlanes = 1;
	bmih->biBitCount = 32;
	bmih->biCompression = BI_RGB;
	bmih->biSizeImage = bmih->biWidth * bmih->biHeight * sizeof(RGBQUAD);
	bmih->biXPelsPerMeter = 0;
   	bmih->biYPelsPerMeter = 0;
	bmih->biClrUsed = 0;
	bmih->biClrImportant = 0;

	DWORD *row = (DWORD *)((BYTE *)texBMI + sizeof(BITMAPINFOHEADER));
	PixelBuf l64(bmw);

	for(int y = 0; y < bmh; y++)	{
		BMM_Color_64 *p64=l64.Ptr();
		// TBD: This should be GetLinearPixels, but since the viewport isn't 
		// yet gamma corrected,  do GetPixels instead.
 
//		int r,g,b;
		newBM->GetLinearPixels(0, y, bmw, p64);
		if (extraFlags&EX_ALPHA_FROM_RGB) {
			BMM_Color_64 *pc = p64;
			for(int x = 0; x < bmw; ++x, ++pc) {
				int a = pc->r+pc->g+pc->b;
				pc->a = a/3;
				}
			}
		if (extraFlags&EX_OPAQUE_ALPHA) {
			BMM_Color_64 *pc = p64;
			for(int x = 0; x < bmw; ++x, ++pc) {
				pc->a = 0xffff;
				}
			}
		if (extraFlags&EX_RGB_FROM_ALPHA) {
			BMM_Color_64 *pc = p64;
			for(int x = 0; x < bmw; ++x, ++pc) {
				USHORT a = pc->a;
				pc->r = a;
				pc->g = a;
				pc->b = a;
				}
			}
		else if (extraFlags&EX_MULT_ALPHA) {
			BMM_Color_64 *pc = p64;
			for(int x = 0; x < bmw; ++x, ++pc) {
				USHORT a = pc->a;
				pc->r = (pc->r*a)>>16;
				pc->g = (pc->g*a)>>16;
				pc->b = (pc->b*a)>>16;
				}
			} 
/*
		for(int x = 0; x < bmw; ++x, ++p64) {
			r = gammaMgr.disp_gamtabw[p64->r>>RCSH16];	
			g = gammaMgr.disp_gamtabw[p64->g>>RCSH16];	
			b = gammaMgr.disp_gamtabw[p64->b>>RCSH16];	
			dw = (r << 16) | (g << 8) | b ; 
			//if ((extraFlags&EX_OPAQUE_ALPHA) || (p64->a & 0x8000)) 
			if (p64->a & 0x8000) 
				dw |= 0xff000000;
			row[x] = dw;
			}
*/
		row += bmih->biWidth;
		}
	return texBMI;
}


BITMAPINFO * ConvertBitmap(Bitmap * bm)
{
	BITMAPINFO* texBMI;

	int newW ;
	int newH ;
	newW = getClosestPowerOf2(bm->Width());
	newH = getClosestPowerOf2(bm->Height());

	if (bm->Width()!=newW||bm->Height()!=newH)
		{ 
		Bitmap *newBM;
		BitmapInfo bi;
		bi.SetName(_T(""));
		bi.SetWidth(newW);
		bi.SetHeight(newH);
		bi.SetType(BMM_TRUE_32);
		bi.SetFlags(bm->HasAlpha() ? MAP_HAS_ALPHA : 0);
		newBM = TheManager->Create(&bi);
		newBM->CopyImage(bm, COPY_IMAGE_RESIZE_LO_QUALITY, 0);
		

		texBMI = BMToDIB(newBM,EX_ALPHA_FROM_RGB);

		newBM->DeleteThis();
		}
	else  {
		texBMI = BMToDIB(bm,0);
		}
	return texBMI;
}

int SimpleZoomCamera::Display(TimeValue t, INode* inode, ViewExp *vpt, int flags) {
	Matrix3 m;
	GraphicsWindow *gw = vpt->getGW();


	INode *node;
	pblock->GetValue(camera_param_basecam,0,node,FOREVER);

	if (node)
		{
		ObjectState os = node->EvalWorldState(0);
		if (os.obj->SuperClassID() == CAMERA_CLASS_ID)
			{
			baseCam = (CameraObject*) os.obj;
			}
		}

	if (initial)
		{
		initial = FALSE;
//get node
		Control *c;
		c = inode->GetTMController();
	
	// Create child node
		nullCtrl = (NullPosition *)	CreateInstance(CTRL_MATRIX3_CLASS_ID,Class_ID(NULL_CONTROL_CLASS_ID,0)); 			

		if (nullCtrl)
			{
			inode->SetTMController(nullCtrl);
//			BOOL iret = c->SetPositionController(nullCtrl);
			nullCtrl->pblock->SetValue(null_params_backpointer,0,(ReferenceTarget*)this);
			}
		}
	else
		{
		nullCtrl = (NullPosition *) inode->GetTMController();

		}


    // a matrix that is viewport scale independent
	Matrix3 tm = inode->GetObjectTM(t);
    gw->setTransform(tm);

    // save the current rendering settings, then set to wireframe
    DWORD rlim = gw->getRndLimits();
    gw->setRndLimits(GW_WIREFRAME|GW_BACKCULL|GW_Z_BUFFER);

    // set an appropriate line colour
     if(inode->Selected())
      gw->setColor(LINE_COLOR,GetSelColor());
    else
    {
      if(!inode->IsFrozen())
        gw->setColor(LINE_COLOR,GetUIColor(COLOR_CAMERA_OBJ));
    }


    // draw the camera itself
    mesh.render(gw,gw->getMaterial(),NULL,COMP_ALL);	


	PBBitmap *bufferPB;		// the zbuffer bitmap param block

	gw->setRndLimits(GW_FLAT|GW_TEXTURE|GW_BACKCULL|GW_Z_BUFFER|GW_ILLUM);
	Material *mat = gw->getMaterial();

	pblock->GetValue(camera_mtl,t,bufferPB,FOREVER);
	if (bufferPB != NULL) 
	{
		bufferPB->Load();
		BITMAPINFO *bitmapInfo =  ConvertBitmap(bufferPB->bm);

		DWORD handle = gw->getTextureHandle(bitmapInfo);
		mat->shadeLimit = GW_FLAT|GW_TEXTURE|GW_BACKCULL|GW_Z_BUFFER|GW_ILLUM;
		mat->texture[0]. textHandle = (DWORD_PTR) &handle;
		
	}



	

//	planeMesh.render(gw,mat,NULL,COMP_ALL,1);	
   
//	delete mat;

	// a matrix that scales with the viewport
//    Matrix3 tm = inode->GetObjectTM(t);
  //  gw->setTransform(tm);


    // restore the rendering settings saved previously
	  gw->setRndLimits(rlim);
  
  return 0;
	}


RefResult SimpleZoomCamera::EvalCameraState(TimeValue t, Interval& valid, CameraState* cs) {
	cs->isOrtho = IsOrtho();	
	cs->fov = GetFOV(t,valid);
	cs->tdist = GetTDist(t,valid);
//	cs->horzLine = horzLineState;
	cs->manualClip = manualClip;
	cs->hither = GetClipDist(t, CAM_HITHER_CLIP, valid);
	cs->yon = GetClipDist(t, CAM_YON_CLIP, valid);
	cs->nearRange = GetEnvRange(t, ENV_NEAR_RANGE, valid);
	cs->farRange = GetEnvRange(t, ENV_FAR_RANGE, valid);
	return REF_SUCCEED;
	}

//
// Reference Managment:
//

RefResult SimpleZoomCamera::NotifyRefChanged(Interval changeInt, RefTargetHandle hTarget, 
	PartID& partID, RefMessage message ) 
	{
	return(REF_SUCCEED);
	}


ObjectState SimpleZoomCamera::Eval(TimeValue time){
	// UpdateUI(time);
	return ObjectState(this);
	}

Interval SimpleZoomCamera::ObjectValidity(TimeValue time) {
	Interval ivalid;
	ivalid.SetInfinite();
		GetFOV(time,ivalid);
		GetTDist(time,ivalid);
		GetClipDist(time, CAM_HITHER_CLIP, ivalid);
		GetClipDist(time, CAM_YON_CLIP, ivalid);
		GetEnvRange(time, ENV_NEAR_RANGE, ivalid);
		GetEnvRange(time, ENV_FAR_RANGE, ivalid);
	return ivalid;	
	}


RefTargetHandle SimpleZoomCamera::Clone(RemapDir& remap) {
	SimpleZoomCamera* newob = new SimpleZoomCamera();
	BaseClone(this, newob, remap);
	return(newob);
	}

#define INITIAL_CHUNK	0x1001

// IO
IOResult SimpleZoomCamera::Save(ISave *isave) 
{
		
	unsigned long nb;

	isave->BeginChunk(INITIAL_CHUNK);
	isave->Write(&initial,sizeof(initial),&nb);
	isave->EndChunk();
	
	return IO_OK;
	}

IOResult  SimpleZoomCamera::Load(ILoad *iload) {
	ULONG nb;
	IOResult res = IO_OK;

	while (IO_OK==(res=iload->OpenChunk())) {
		switch (iload->CurChunkID()) {
			case INITIAL_CHUNK:
				iload->Read(&initial,sizeof(initial),&nb);
				break;

			}		
		iload->CloseChunk();
		if (res!=IO_OK)  return res;
		}
	return IO_OK;
	}

