/**********************************************************************
 *<
	FILE: camera.h

	DESCRIPTION:  Defines a simple Camera

	CREATED BY: Dan Silva

	HISTORY: created 13 September 1994

 *>	Copyright (c) 1994, All Rights Reserved.
 **********************************************************************/

#ifndef __CAMERA__H__ 

#define __CAMERA__H__


#define ZOOMSIMPLE_CAM_CLASS_ID 0xee1001

#include "iparamm2.h"
#include "Simpobj.h"

// reference #s
#define PBLOCK_REF		0

#define NULL_CONTROL_CLASS_ID 0x1256788a

// main block
enum { null_params};
// path_params param IDs
enum { null_params_backpointer,null_param_zoom,null_param_panx,null_param_pany,
		null_param_zoffset,
		null_param_animatable};



// block IDs
enum { camera_param};

enum { camera_param_basecam, camera_param_zoom, camera_param_panx,camera_param_pany,
		camera_mtl };


#define NULL_CONTROL_CNAME		GetString(IDS_PW_NULL)


class NullPosition : public Control {
	public:
		int axisConstraint, coordSys;
		float tval;
		float sx,sy;
		int mode;
		IParamBlock2* pblock;
		Point3 curval;
		Interval ivalid;
		
		static NullPosition *editCont;
		static IObjParam *ip;

		NullPosition(const NullPosition &ctrl);
		NullPosition(BOOL loading=FALSE);
		~NullPosition();

		void ResetParams();
		
		void Update(TimeValue t);

		// Animatable methods
		Class_ID ClassID() { return Class_ID(NULL_CONTROL_CLASS_ID,0); }  
		SClass_ID SuperClassID() { return CTRL_MATRIX3_CLASS_ID; }  		
		
		void GetClassName(TSTR& s) {s = NULL_CONTROL_CNAME;}
		void DeleteThis() {delete this;}		
		int IsKeyable() {return 0;}		

		int NumSubs()  {return 1;}
		Animatable* SubAnim(int i) {return (Animatable *) pblock;}
		TSTR SubAnimName(int i) { return GetString(IDS_PW_NULLPARAMS); }
		int SubNumToRefNum(int subNum) {if (subNum==0) return PATHPOS_PBLOCK_REF; else return -1;}

		void BeginEditParams( IObjParam *ip, ULONG flags,Animatable *prev ); 
		void EndEditParams( IObjParam *ip, ULONG flags,Animatable *next ); 

//		int SetProperty(ULONG id, void *data);
//		void *GetProperty(ULONG id);

// JBW: direct ParamBlock access is added
		int	NumParamBlocks() { return 1; }					// return number of ParamBlocks in this instance
		IParamBlock2* GetParamBlock(int i) { return pblock; } // return i'th ParamBlock
		IParamBlock2* GetParamBlockByID(BlockID id) { return (pblock->ID() == id) ? pblock : NULL; } // return id'd ParamBlock

		// Reference methods
		int NumRefs() { return 1; };	
		RefTargetHandle GetReference(int i);
		void SetReference(int i, RefTargetHandle rtarg);
		RefResult NotifyRefChanged(Interval, RefTargetHandle, PartID&, RefMessage);

//		IOResult Save(ISave *isave);
//		IOResult Load(ILoad *iload);

		void RescaleWorldUnits(float f) {}

		// Control methods
		void Copy(Control *from);
		RefTargetHandle Clone(RemapDir& remap);
		BOOL IsLeaf() {return TRUE;}
		void GetValue(TimeValue t, void *val, Interval &valid, GetSetMethod method);	
		void SetValue(TimeValue t, void *val, int commit, GetSetMethod method);
		void CommitValue(TimeValue t) { }
		void RestoreValue(TimeValue t) { }

		int complete;
		void MouseCycleStarted(TimeValue t);
		void MouseCycleCompleted(TimeValue t);
		Matrix3 objectTM;
	};




class SimpleZoomCamera : public CameraObject
{
	friend INT_PTR_MSVC70 CALLBACK SimpleCamParamDialogProc( HWND hDlg, UINT message,
		 WPARAM wParam, LPARAM lParam );

	CameraObject *baseCam;
	NullPosition *nullCtrl;

	BOOL initial;

	// Class vars
	static Mesh mesh;
	Mesh planeMesh;
	static short meshBuilt;
	static BOOL inCreate;

	static HWND hSimpleCamParams;
	static IObjParam* iObjParams;		
	

// JBW: direct ParamBlock access is added
	int	NumParamBlocks() { return 1; }					// return number of ParamBlocks in this instance
	IParamBlock2* GetParamBlock(int i) {	if (i==0) return pblock;
											return NULL;} // return i'th ParamBlock
	IParamBlock2* GetParamBlockByID(BlockID id) { if (pblock->ID() == id) return pblock;
												else return  NULL; } // return id'd ParamBlock


	//  inherited virtual methods for Reference-management
	RefResult NotifyRefChanged( Interval changeInt, RefTargetHandle hTarget, PartID& partID, RefMessage message );
	void BuildMesh();




public:

	IParamBlock2 *pblock;

	void ResetParams();

	int manualClip;
	BOOL ortho;
	float  rangeDisplay;

	SimpleZoomCamera(int look = 0);
	~SimpleZoomCamera();

	// From Camera
	RefResult EvalCameraState(TimeValue t, Interval& valid, CameraState* cs);
	void SetFOV(TimeValue t, float f);
	float GetFOV(TimeValue t, Interval& valid = Interval(0,0));
	void SetTDist(TimeValue t, float f);
	float GetTDist(TimeValue t, Interval& valid = Interval(0,0));

	void SetOrtho(BOOL b) { ortho = b;};
	BOOL IsOrtho() {return FALSE;};

	int GetManualClip() { return manualClip; }
	void SetManualClip(int onOff);
	float GetClipDist(TimeValue t, int which, Interval &valid = Interval(0,0));
	void SetClipDist(TimeValue t, int which, float f);
	void SetEnvRange(TimeValue time, int which, float f);
	float GetEnvRange(TimeValue t, int which, Interval& valid = Interval(0,0));
	void SetEnvDisplay(BOOL b, int notify=TRUE);
	BOOL GetEnvDisplay(void)	{ return rangeDisplay; }
	void RenderApertureChanged(TimeValue t);
	void UpdateTargDistance(TimeValue t, INode* inode);

	//  inherited virtual methods:

	// From BaseObject
	int HitTest(TimeValue t, INode* inode, int type, int crossing, int flags, IPoint2 *p, ViewExp *vpt);
	void Snap(TimeValue t, INode* inode, SnapInfo *snap, IPoint2 *p, ViewExp *vpt);
	int Display(TimeValue t, INode* inode, ViewExp *vpt, int flags);
	CreateMouseCallBack* GetCreateMouseCallBack();
	void BeginEditParams( IObjParam *ip, ULONG flags,Animatable *prev);
	void EndEditParams( IObjParam *ip, ULONG flags,Animatable *next);
	TCHAR *GetObjectName() { return  GetString(IDS_DB_FREE_CAM); }
	int Type() { return FREE_CAMERA; }
	void SetType(int tp);
	void GetWorldBoundBox(TimeValue t, INode *mat, ViewExp* vpt, Box3& box);
	void GetLocalBoundBox(TimeValue t, INode *mat, ViewExp* vpt, Box3& box);
	void GetDeformBBox(TimeValue t, Box3& box, Matrix3 *tm, BOOL useSel);

	// From Object
	ObjectState Eval(TimeValue time);
	void InitNodeName(TSTR& s) { s = GetString(IDS_DB_CAMERA); }
	Interval ObjectValidity();
	int DoOwnSelectHilite() { return 1; }
	Interval ObjectValidity(TimeValue time);
	

	// Animatable methods
	void DeleteThis() { delete this; }
	Class_ID ClassID() { return Class_ID(ZOOMSIMPLE_CAM_CLASS_ID, 0); }
	void GetClassName(TSTR& s) { s = TSTR( GetString(IDS_DB_FREE_CLASS) ); }
	
	int NumSubs()
	{
		return 1;
	}
	Animatable* SubAnim(int i)
	{
		switch(i)
		{
			case PBLOCK_REF:
				return (Animatable*)pblock;


			default:
				return NULL;
		}
	}

	TSTR SubAnimName(int i)
	{
		switch(i)
		{
			case PBLOCK_REF:
				return TSTR( GetString(IDS_RB_PARAMETERS) );

			default:
				return _T("");
		}
	}

	// From ref
	RefTargetHandle Clone( RemapDir& remap = NoRemap() );

	int NumRefs() { return 1; } // mjm - 07.17.00

	RefTargetHandle GetReference(int i)
	{
		switch(i)
		{
			case PBLOCK_REF:
				return (RefTargetHandle)pblock;


			default:
				return NULL;
		}
	}

	void SetReference(int i, RefTargetHandle rtarg)
	{
		switch(i)
		{
			case PBLOCK_REF:
				pblock=(IParamBlock2*)rtarg;
				break;


			default:
				break;
		}
	}

	// IO
	IOResult Save(ISave *isave);
	IOResult Load(ILoad *iload);

};


#define GRIDID 0x81f1dfc, 0x77566f88
#define A_RENDER			A_PLUGIN1
#define PBLOCK_REF_NO	 0

class GridObject : public SimpleObject2{
	public:
		// Class vars
		static IObjParam *ip;
		int createMeth;
		Point3 crtPos;		
		static float crtWidth, crtHeight, crtLength;
		static BOOL typeinCreate;
		float RenderWidth,RenderLength;

		GridObject();
		
		// From Object
		int CanConvertToType(Class_ID obtype);
		void GetCollapseTypes(Tab<Class_ID> &clist,Tab<TSTR*> &nlist);
		Object* ConvertToType(TimeValue t, Class_ID obtype);
		
		// From BaseObject
		CreateMouseCallBack* GetCreateMouseCallBack();
		void BeginEditParams( IObjParam *ip, ULONG flags,Animatable *prev);
		void EndEditParams( IObjParam *ip, ULONG flags,Animatable *next);
		TCHAR *GetObjectName() { return GetString(IDS_PW_GRID_OBJECT); }

// JBW: direct ParamBlock access is added
		int	NumParamBlocks() { return 1; }					// return number of ParamBlocks in this instance
		IParamBlock2* GetParamBlock(int i) { return pblock2; } // return i'th ParamBlock
		IParamBlock2* GetParamBlockByID(BlockID id) { return (pblock2->ID() == id) ? pblock2 : NULL; } // return id'd ParamBlock


		// Animatable methods
		void DeleteThis() { delete this; }
		Class_ID ClassID() { return Class_ID( GRIDID); }  
		
		// From ref
		RefTargetHandle Clone(RemapDir& remap = NoRemap());
		IOResult Load(ILoad *iload);

		// From SimpleObject
		void BuildMesh(TimeValue t);
		void BuildGridPatch(TimeValue t, PatchMesh &patch);


		BOOL OKtoDisplay(TimeValue t);
		void InvalidateUI();
		BOOL HasUVW() ;
		void SetGenUVW(BOOL sw);

		int RenderBegin(TimeValue t, ULONG flags);
		int RenderEnd(TimeValue t);

	};				


//
//	Creation method grid_creation_type,


enum {  grid_creation_type,grid_type_in, grid_params };
// grid_creation_type param IDs
enum { grid_create_meth };
// grid_type_in param IDs
enum { grid_ti_pos, grid_ti_length, grid_ti_width};
// grid_param param IDs
enum { grid_length,
	   grid_width,
	   grid_wsegs,
	   grid_lsegs,
	   grid_rsegs,
	   grid_rscale,
	   grid_genuvs,
	   grid_leftover,
	   grid_leftover1,
	   grid_camera,
	   grid_zdepth,
	   grid_hoscan, grid_woscan,
	   grid_centerx,grid_centery,

	   grid_panx, grid_pany,
	   grid_fov,
	   grid_animatable

	 };




#endif
