/**********************************************************************
 *<
	FILE: pathctrl.cpp

	DESCRIPTION: A controller that moves an object along a spline path
				 Ed. 2 re-coded using ParamBlock2/ParamMap2

	CREATED BY: Rolf Berteig
	            Ed. 2 John Wainwright

	HISTORY: created 13 June 1995
	         Ed. 2 re-coded 9/22/98

 *>	Copyright (c) 1994, All Rights Reserved.
 **********************************************************************/

//#include "prim.h"
//#include "units.h"
//#include "interpik.h"
//#include "spline3d.h"
#include "zoomcam.h"
#include "camera.h"
#include "istdplug.h"
#include "iparamm2.h"

IObjParam *NullPosition::ip             = NULL;
NullPosition *NullPosition::editCont    = NULL;

//********************************************************
// PATH CONTROL
//********************************************************
static Class_ID pathControlClassID(NULL_CONTROL_CLASS_ID,0); 
class NullClassDesc:public ClassDesc2 {
	public:
	int 			IsPublic() { return 0; }
	void *			Create(BOOL loading) { return new NullPosition(loading); }
	const TCHAR *	ClassName() { return NULL_CONTROL_CNAME; }
	SClass_ID		SuperClassID() { return CTRL_MATRIX3_CLASS_ID; }
	Class_ID		ClassID() { return Class_ID(NULL_CONTROL_CLASS_ID,0); }
	const TCHAR* 	Category() { return _T("");  }
// JBW: new descriptor data accessors added.
	const TCHAR*	InternalName() { return _T("Null"); }			// returns fixed parsable name (scripter-visible name)
	HINSTANCE		HInstance() { return hInstance; }				// returns owning module handle
	};

static NullClassDesc nullCD;
ClassDesc* GetNullCtrlDesc() {return &nullCD;}

// ParamBlock IDs

// per instance path controller block
static ParamBlockDesc2 null_paramblk (null_params, _T("NullParameters"),  0, &nullCD, P_AUTO_CONSTRUCT + P_AUTO_UI, 0, 
	//rollout
	IDD_NULL, IDS_RB_PARAMETERS, BEGIN_EDIT_MOTION, 0, NULL,
	null_params_backpointer,	_T("backpointer"),   TYPE_REFTARG,	0,	IDS_PW_BACK,
		end,

	null_param_zoom,  _T("zoom"), 			TYPE_FLOAT, 	P_ANIMATABLE + P_RESET_DEFAULT, 	IDS_PW_ZOOM, 
		p_default, 		1.0f,	
		p_ms_default,	1.0f,
		p_range, 		0.01f, 50.f, 
		p_ui, 			TYPE_SPINNER, EDITTYPE_UNIVERSE, IDC_ZOOM, IDC_ZOOMSPINNER, 0.02f, 
		end, 

	null_param_panx,  _T("pan_x"), 			TYPE_FLOAT, 	P_ANIMATABLE + P_RESET_DEFAULT, 	IDS_PW_PANX, 
		p_default, 		0.0f,	
		p_ms_default,	0.0f,
		p_range, 		-100.0f, 100.0f, 
		p_ui, 			TYPE_SPINNER, EDITTYPE_UNIVERSE, IDC_PANX, IDC_PANXSPINNER, 0.01f, 
		end, 

	null_param_pany,  _T("pan_y"), 			TYPE_FLOAT, 	P_ANIMATABLE + P_RESET_DEFAULT, 	IDS_PW_PANY, 
		p_default, 		0.0f,	
		p_ms_default,	0.0f,
		p_range, 		-100.0f, 100.0f, 
		p_ui, 			TYPE_SPINNER, EDITTYPE_UNIVERSE, IDC_PANY, IDC_PANYSPINNER, 0.01f, 
		end, 

	null_param_zoffset,  _T("offset_z"), 			TYPE_FLOAT, 	P_ANIMATABLE + P_RESET_DEFAULT, 	IDS_PW_OFFSETZ, 
		p_default, 		0.0f,	
		p_ms_default,	0.0f,
		p_range, 		-100000.0f, 100000.0f, 
		p_ui, 			TYPE_SPINNER, EDITTYPE_UNIVERSE, IDC_ZOFFSET, IDC_ZOFFSETSPINNER, 0.01f, 
		end, 

	null_param_animatable, 	_T("animatablePanZoom"), 	TYPE_BOOL, 		0,				IDS_PW_ANIMATABLE,
		p_default, 		FALSE, 
		p_ui, 			TYPE_SINGLECHEKBOX, 	IDC_ANIMATABLE_CHECK, 
		end, 



	end
	);


class NullDlgProc : public ParamMap2UserDlgProc {
public:
	NullPosition *mod;		
	NullDlgProc(NullPosition *m) {mod = m;}		
	BOOL DlgProc(TimeValue t,IParamMap2 *map,HWND hWnd,UINT msg,WPARAM wParam,LPARAM lParam);		
	void DeleteThis() {delete this;}		
};

BOOL NullDlgProc::DlgProc (TimeValue t,IParamMap2 *map,HWND hWnd,UINT msg,WPARAM wParam,LPARAM lParam)
	{
	switch (msg) {		
		case WM_COMMAND:
			switch (LOWORD(wParam)) {

				case IDC_RESET:
					mod->ResetParams();
					break;

		
				}
			break;
		}
	return FALSE;
	}


NullPosition::NullPosition(BOOL loading) 
	{
	curval   = Point3(0,0,0);
	// make the paramblock
	nullCD.MakeAutoParamBlocks(this);
	sx = 0.0f;
	sy = 0.0f;
}

NullPosition::~NullPosition()
	{
	DeleteAllRefsFromMe();
	}

RefTargetHandle NullPosition::Clone(RemapDir& remap)
	{
	NullPosition *p = new NullPosition(TRUE);

    p->ReplaceReference(0, pblock->Clone(remap));
	return p;
	}


void NullPosition::Copy(Control *from)
	{
	if (from->ClassID()==ClassID()) {
		NullPosition *ctrl = (NullPosition*)from;
		RemapDir *remap = NewRemapDir(); 
		ReplaceReference(0, ctrl->pblock->Clone(*remap));
		remap->DeleteThis();
	} 
	ivalid.SetEmpty();
	NotifyDependents(FOREVER,PART_ALL,REFMSG_CHANGE);
	}


void NullPosition::Update(TimeValue t)
	{
	}


void NullPosition::MouseCycleStarted(TimeValue t)
{
}


static BOOL refStop = FALSE;

void NullPosition::MouseCycleCompleted(TimeValue t)
{

}


void NullPosition::ResetParams()
	{
	ReferenceTarget *ref;
	Interval iv;
	pblock->GetValue(null_params_backpointer,0,ref,iv);

	SimpleZoomCamera *ob = (SimpleZoomCamera *) ref;

//	INode *node;
//	ob->pblock->GetValue(camera_param_basecam,0,node,FOREVER);

	if (ob)
		{
		float zoom = 1.0f;
		if (ob) zoom = ob->GetFOV(0,FOREVER);
		pblock->SetValue(null_param_zoom,0,zoom);
		pblock->SetValue(null_param_panx,0,0.0f);
		pblock->SetValue(null_param_pany,0,0.0f);
		}
	}



void NullPosition::SetValue(TimeValue t, void *val, int commit, GetSetMethod method)
{

	SetXFormPacket *ptr = (SetXFormPacket*)val;
	
	// RB 8/14/2000
//	Matrix3 ptm = ApplyInheritance(t,ptr->tmParent,pos);	

	if (ptr->command == XFORM_MOVE)
		{
		int cid = GetCOREInterface()->GetCommandMode()->Class();
		int id = GetCOREInterface()->GetCommandMode()->ID();
//		if ( id == CID_CAMTRUCK)
			{
//			refStop = TRUE;
			pblock->GetValue(null_param_panx,t,sx,FOREVER);
			pblock->GetValue(null_param_pany,t,sy,FOREVER);
			float zoom;
			pblock->GetValue(null_param_pany,t,zoom,FOREVER);

			BOOL animatable;
			pblock->GetValue(null_param_animatable,t,animatable,FOREVER);

			sx -= ptr->p.x/500.f;//*zoom;
			sy -= ptr->p.y/500.f;//*zoom;

			if (!animatable)
			{
				SuspendAnimate();
				AnimateOff();
			}

			pblock->SetValue(null_param_panx,t,sx);
			pblock->SetValue(null_param_pany,t,sy);

			if (!animatable)
			{
				ResumeAnimate();				
			}


//			refStop = FALSE;
//			Move(t,ptr->tmParent,ptr->tmAxis,ptr->p,ptr->localOrigin,commit);
			}
		}

/*
if (method==CTRL_RELATIVE) 
	{
	Point3 *inVal = ((Point3 *) val);
	Point3 inputVec = *inVal;

	Interval valid;
	ReferenceTarget *ref;
	Interval iv;
	pblock->GetValue(null_params_backpointer,0,ref,iv);
	Float2dObject *ob = (Float2dObject *) ref;
//need to put everything in screen space then ignore the z component
	ViewExp *vExp = GetCOREInterface()->GetActiveViewport();
	Matrix3 worldToScreenTm;
	vExp->GetAffineTM(worldToScreenTm);
	Point3 zero(0.0f,0.0f,0.0f);
	float sc = vExp->GetVPWorldWidth( zero);

	GraphicsWindow *gw = vExp->getGW();
	float swidth  = (float)gw->getWinSizeX();
	float sheight  = (float)gw->getWinSizeY();
	DebugPrint("%f input Vec %f %f %f\n", sc,inputVec.x,inputVec.y, inputVec.z);
	inputVec = VectorTransform(worldToScreenTm,inputVec);




	float width,min,max;
	ob->pblock2->GetValue(float2d_width,t,width,iv);
	ob->pblock2->GetValue(float2d_min,t,min,iv);
	ob->pblock2->GetValue(float2d_max,t,max,iv);
	if (mode == 2)
		{
//DebugPrint("float %f\n",inputVec.x);
		inputVec.x = inputVec.x/sc * swidth * 0.75f;
		inputVec.y = inputVec.y/sc * sheight * 0.75f;
		inputVec.x /= width;
		inputVec.x *= (max-min);
		float newVal = tval+inputVec.x;
		
		ob->pblock2->SetValue(float2d_value,t,newVal);
		}
	else if (mode == 0)
		{
		inputVec.x = inputVec.x/sc  * 0.75f;
		inputVec.y = inputVec.y/sc  * 1.f;
		float newVal = sx+inputVec.x;
		if (newVal < 0.01f) newVal = 0.01f;
		ob->pblock2->SetValue(float2d_posx,t,newVal);
		newVal = sy-inputVec.y;
		if (newVal < 0.05f) newVal = 0.05f;
		if (newVal > 0.95f) newVal = 0.95f;
		ob->pblock2->SetValue(float2d_posy,t,newVal);
		}


	}
*/
 
}


void NullPosition::GetValue(TimeValue t, void *val, Interval &valid, GetSetMethod method)
	{
//		Interval valid;
		ReferenceTarget *ref;
		Interval iv;

		float panx,pany;

		pblock->GetValue(null_params_backpointer,0,ref,iv);
		pblock->GetValue(null_param_panx,t,panx,iv);
		pblock->GetValue(null_param_pany,t,pany,iv);

		float zlen = 0.0f;
		pblock->GetValue(null_param_zoffset,t,zlen,iv);
//		panx += sx;
//		pany += sy;

		SimpleZoomCamera *ob = (SimpleZoomCamera *) ref;

		INode *node;
		ob->pblock->GetValue(camera_param_basecam,0,node,FOREVER);

		if (node)
			{
			Matrix3 *mat = (Matrix3*)val;
			Matrix3 tm = node->GetObjectTM(t);
//now manipulate the tm to skew it
			Point3 x = Normalize(tm.GetRow(0));
			Point3 y = Normalize(tm.GetRow(1));
			Point3 z = tm.GetRow(2) + (x*panx) + (y*pany);

			Point3 zvec = Normalize(tm.GetRow(2)) * zlen;
//			*mat = tm;
			mat->Set(x,y,z,tm.GetRow(3)+zvec);
			}
		valid.SetInstant(t);

  		

	}

RefTargetHandle NullPosition::GetReference(int i)
	{
	return pblock;
	}

void NullPosition::SetReference(int i, RefTargetHandle rtarg)
	{
	pblock = (IParamBlock2*)rtarg; 
	}

RefResult NullPosition::NotifyRefChanged(
		Interval iv, 
		RefTargetHandle hTarg, 
		PartID& partID, 
		RefMessage msg) 
	{
	if (refStop) return REF_STOP;
	switch (msg) {
		case REFMSG_CHANGE:
			ivalid.SetEmpty();
			null_paramblk.InvalidateUI();
			break;
		}

	return REF_SUCCEED;
	}

/*--------------------------------------------------------------------*/
// NullPosition UI

void NullPosition::BeginEditParams( IObjParam *ip, ULONG flags,Animatable *prev )
	{	
	this->ip = ip;
	editCont = this;
//	if (flags & BEGIN_EDIT_HIERARCHY) {
// JBW: JointParam stuff not handled by ParamBlock2 yet		
		// No IK if follow is on
		nullCD.BeginEditParams(ip, this, flags, prev);
		null_paramblk.SetUserDlgProc(new NullDlgProc(this));

//		}
	}

void NullPosition::EndEditParams( IObjParam *ip, ULONG flags,Animatable *next )
{
	editCont = NULL;
	IParamMap2* pmap = pblock->GetMap();
	nullCD.EndEditParams(ip, this, flags | END_EDIT_REMOVEUI, next);
}


/*
#define INITIAL_CHUNK	0x1001

IOResult NullPosition::Save(ISave *isave)
	{	
	isave->BeginChunk(JOINTPARAMPATH_CHUNK);
	isave->Write(&initial,sizeof(initial),&nb);
	isave->EndChunk();
	
	return IO_OK;
	}

IOResult NullPosition::Load(ILoad *iload)
	{
	ULONG nb;
	IOResult res = IO_OK;

	while (IO_OK==(res=iload->OpenChunk())) {
		switch (iload->CurChunkID()) {
			case INITIAL_CHUNK:
				iload->Read(&initial,sizeof(initial),&nb);
				break;

			}		
		iload->CloseChunk();
		if (res!=IO_OK)  return res;
		}
	return IO_OK;
	}
*/
